@@ -5,9 +5,8 @@ use warnings;
 
 use Carp ();
 
-our $VERSION = '0.30';
+our $VERSION = '0.31';
 our $PREFIX  = '__cag_';
-our $NEED_COMPAT = 0;
 
 require XSLoader;
 XSLoader::load('Class::Accessor::Inherited::XS', $VERSION);
@@ -33,7 +32,6 @@ sub import {
 
     my %opts = ref($_[0]) eq 'HASH' ? %{ $_[0] } : @_;
     my $class = delete $opts{package} // caller;
-    local $NEED_COMPAT = 1 if delete $opts{compability};
 
     for my $type (keys %opts) {
         my $accessors = $opts{$type};
@@ -108,7 +106,7 @@ sub register_type {
         $args->{installer} = sub {
             my ($class, $name, $field) = @_;
             install_inherited_cb_accessor(
-                "${class}::${name}", $field, $PREFIX.$field, $args->{read_cb} // $args->{on_read}, $args->{write_cb} // $args->{on_write}
+                "${class}::${name}", $field, $PREFIX.$field, $args->{read_cb} // $args->{on_read}, $args->{write_cb} // $args->{on_write}, 0
             );
         };
     }
@@ -131,13 +129,13 @@ sub _type_installer {
 sub _mk_inherited_accessor {
     my ($class, $name, $field, $flags) = @_;
 
-    install_inherited_accessor("${class}::${name}", $field, $PREFIX.$field, $flags | ($NEED_COMPAT ? 2 : 0));
+    install_inherited_accessor("${class}::${name}", $field, $PREFIX.$field, $flags);
 }
 
 sub _mk_class_accessor {
-    my ($class, $name, $default, $is_varclass, $is_readonly) = @_;
+    my ($class, $name, $default, $is_varclass, $is_readonly, $is_weak) = @_;
 
-    install_class_accessor("${class}::${name}", $default, $is_varclass, $is_readonly);
+    install_class_accessor("${class}::${name}", $default, $is_varclass, ($is_readonly ? 1 : 0) | ($is_weak ? 2 : 0));
 }
 
 sub _mk_object_accessor {
@@ -329,7 +327,7 @@ and $__PACKAGE__::accessor variable. While the stored value is still accessible
 
 Copyright (C) 2009 by Vladimir Timofeev
 
-Copyright (C) 2014-2015 by Sergey Aleynikov
+Copyright (C) 2014-2016 by Sergey Aleynikov
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.10.1 or,
